/* bbowda - Black Box Optimization With Data Analysis
   Copyright (C) 2006-2007 Kevin Kofler <Kevin@tigcc.ticalc.org>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version. A copy of the GNU General Public
   License version 3 can be found in the file gpl-3.0.txt.

   Linking bbowda statically or dynamically (directly or indirectly) with
   other modules is making a combined work based on bbowda. Thus, the terms
   and conditions of the GNU General Public License cover the whole
   combination.

   In addition, as a special exception, the copyright holder of bbowda gives
   you permission to combine the bbowda program:
   * with free software programs or libraries that are released under the
     GNU Library or Lesser General Public License (LGPL), either version 2
     of the License, or (at your option) any later version,
   * with free software programs or libraries that are released under the
     IBM Common Public License (CPL), either version 1.0 of the License, or
     (at your option) any later version,
   * with code included in the standard release of MUMPS under the MUMPS
     Conditions of Use as reproduced in licenses.txt (or modified versions
     of such code, with unchanged license) and
   * if you qualify for a free of charge license of DONLP2, with code
     included in the standard release of DONLP2 under the DONLP2 Conditions
     of Use as reproduced in licenses.txt (or modified versions of such code,
     with unchanged license).

   You may copy and distribute such a system following the terms of the GNU
   GPL for bbowda and the licenses of the other code concerned, provided that
   you include the source code of that other code when and as the GNU GPL
   requires distribution of source code.

   Note that people who make modified versions of bbowda are not obligated
   to grant this special exception for their modified versions; it is their
   choice whether to do so. The GNU General Public License gives permission
   to release a modified version without this exception; this exception also
   makes it possible to release a modified version which carries forward
   this exception.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define _ISOC99_SOURCE
#include "nlpopt.h"
#include "eval.h"
#include "gmmem.h"
#include "eqconst.h"
#include "covar.h"
#include "problem.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* **************************************************************************** */
/*                                 main function                                */
/* **************************************************************************** */

static const int enable_global_search=TRUE;

int main(void) {
  size_t i;
  int ret;

  /* ensure reproducible results */
  srand(31337);

  /* evaluate constraints at starting points */
  init_points();
  /* if we don't have enough starting points, generate more */
  #if DIMX<5
    #define N_START (DIMX+DIMY+DIMY_EQ+6)
  #elif DIMX<10
    #define N_START (DIMX+DIMX+DIMY+DIMY_EQ+2)
  #else
    #define N_START (4*DIMX+DIMY+DIMY_EQ-16)
  #endif
  if (numcurrpts < N_START) {
    /* consider a grid of (N_START*2)*(N_START*2) points
       don't generate 2 points on the same xk */
    unsigned char used_x[DIMX][N_START*2]={};
    /* mark all the already used xk */
    size_t i, k, n_used_x[DIMX]={};
    DOUBLE delta_x[DIMX];
    for (k=0; k<(size_t)DIMX; k++) {
      delta_x[k]=(xup[k]-xlow[k])/(DOUBLE)(N_START*2-1);
    }
    for (i=0; i<numcurrpts; i++) {
      /* round the given starting points to the closest x and y on the grid */
      DOUBLE *x=currpts[i];
      for (k=0; k<(size_t)DIMX; k++) {
        if (delta_x[k]!=0.) { /* filter out equality bounds */
          long j=lround((x[k]-xlow[k])/delta_x[k]);
          if (j<0) { /* sanity check */
            printf("WARNING: Starting value %lf out of bounds [%lf,%lf].\n",
                   x[k],xlow[k],xup[k]);
            j=0;
          }
          if (j>=2*N_START) { /* sanity check */
            printf("WARNING: Starting value %lf out of bounds [%lf,%lf].\n",
                   x[k],xlow[k],xup[k]);
            j=2*N_START-1;
          }
          if (!used_x[k][j]) n_used_x[k]++;
          used_x[k][j]=TRUE;
        }
      }
    }
    /* now loop until we have enough starting points */
    while (numcurrpts < N_START) {
      /* generate a starting point */
      /* try 10 points, pick the one the farthest away */
      DOUBLE new_x[DIMX], dist=-INFINITY;
      size_t t, j[DIMX];
      for (t=0; t<10; t++) {
        DOUBLE new_x_t[DIMX], dist_t=INFINITY;
        size_t j_t[DIMX], i;
        /* generate point */
        for (k=0; k<(size_t)DIMX; k++) {
          /* equiprobability among the not yet used xk */
          size_t j,jtk_index=((unsigned long long)rand()
                               *(unsigned long long)((size_t)(N_START*2)-n_used_x[k]))
                             /((unsigned long long)(RAND_MAX)+1ull);
          for (j=0; j<(size_t)(N_START*2); j++) {
            if (!used_x[k][j]) {
              if (!jtk_index--) {
                new_x_t[k]=xlow[k]+(DOUBLE)j*delta_x[k];
                j_t[k]=j;
                break;
              }
            }
          }
        }
        /* compute squared distance to the closest point */
        for (i=0; i<numcurrpts; i++) {
          DOUBLE *x=currpts[i], dist_t_i=0.;
          for (k=0; k<(size_t)DIMX; k++) {
            dist_t_i+=(new_x_t[k]-x[k])*(new_x_t[k]-x[k]);
          }
          if (dist_t_i<dist_t) dist_t=dist_t_i;
        }
        /* keep the point if it's better */
        if (dist_t>dist) {
          dist=dist_t;
          memcpy(j,j_t,(size_t)DIMX*sizeof(size_t));
          memcpy(new_x,new_x_t,(size_t)DIMX*sizeof(DOUBLE));
        }
      }
      /* mark the xk[j] as used */
      for (k=0; k<(size_t)DIMX; k++) {
        used_x[k][j[k]]=TRUE;
        n_used_x[k]++;
      }
      /* evaluate constraints at the new starting point */
      printf("Generated starting point x=[");
      if (DIMX) {
        printf("%lf",new_x[0]);
      }
      for (i=1; i<(size_t)DIMX; i++) {
        printf(",%lf",new_x[i]);
      }
      printf("]\n");
      new_point(new_x);
    }
  }

  /* compute first global over-/underestimates of equality constraints */
  compute_1st_global_eq_cst_estimates();

  /* optimize until we have reached the maximum number of evaluations */
  while (numcurrpts < MAXPTS) {
    if (enable_global_search && (numcurrpts & 1)) { /* global (gapfilling) search */
      /* compute a density function (gmm) */
      build_density_gmm();
      /* minimize the density function under the constraints */
      ignore_constraints=TRUE;
      if (num_estimate_constraints) {
        ret=solve_nlp();
        printf("Global presearch found x=[");
        if (DIMX) {
          printf("%lf",optimum_x[0]);
        }
        for (i=1; i<(size_t)DIMX; i++) {
          printf(",%lf",optimum_x[i]);
        }
        printf("] (reason: %d)\n",ret);
        ignore_constraints=FALSE;
      }
      ret=solve_nlp();
      /* free the density function parameters */
      free_gmm();
      printf("Global search found x=[");
    } else { /* local search */
      /* choose best point */
      get_best_point();
      /* build regularized covariance model around the point */
      build_local_regcovar_model();
      /* optimize the covariance model */
      ret=solve_nlp();
      printf("Local search found x=[");
    }
    if (DIMX) {
      printf("%lf",optimum_x[0]);
    }
    for (i=1; i<(size_t)DIMX; i++) {
      printf(",%lf",optimum_x[i]);
    }
    printf("] (reason: %d)\n",ret);
    /* evaluate constraints at optimum */
    new_point(optimum_x);
  }

#if DIMY_EQ
  extrapolate_point();
#endif

  printf("Optimum: %lf at x=[",get_optimum_x());
  if (DIMX) {
    printf("%lf",optimum_x[0]);
  }
  for (i=1; i<(size_t)DIMX; i++) {
    printf(",%lf",optimum_x[i]);
  }
  printf("]\n");

  return 0;
}
